
from blocks import Block

class Node:
	successors = None
	predecessors = None
	block = None
	doms = None
	s_doms = None
	df = None
	isStart = None
	isEnd = None

	def __init__(self):
		self.successors = []
		self.predecessors = []
		self.doms = set()
		self.s_doms = set()
		self.df = set()
		self.isStart = False
		self.isEnd = False
		self.block = Block()
	
	def toString(self):
		if self.isStart:
			return "START"
		if self.isEnd:
			return "END"
		else:
			return str(self.block.instructions[0].lineNumber)
		

class CFG:
	nodes = None
	startnode = None
	endnode = None

	def __init__(self):
		self.nodes = []
		self.startnode = Node()
		self.startnode.isStart = True
		self.endnode = Node()
		self.endnode.isEnd = True
	
	def genDomSets(self):
		# initialize dominated by sets
		for n in self.nodes:
			if n == self.startnode:
				n.domby = set([self.startnode])
			else:
				n.domby = set(self.nodes)

		# build dominated by set
		world = set(self.nodes)
		world.remove(self.startnode)
		changed = True
		while changed:
			changed = False
			for n in world:
				domby = reduce(set.__and__, (m.domby for m in n.predecessors)).union(set([n]))
				if n.domby != domby:
					changed = True
					n.domby = domby

		# build doms set
		for n in self.nodes:
			for m in self.nodes:
				if n in m.domby:
					n.doms.add(m)

		# build s_doms set
		for n in self.nodes:
			n.s_doms = n.doms - set([n])

		# build df set
		for x in self.nodes:
			for y in self.nodes:
				for p in y.predecessors:
					if p in x.doms and y not in x.s_doms:
						x.df.add(y)

		# test doms set
		#for n in self.nodes:
		#	print n.toString() + "\tdf is:\t",
		#	for d in n.df:
		#		print d.toString() + " ",
		#	print ""
		#print ""

	def help_makeCFG(self):
		# connect the nodes
		for n1 in self.nodes: # for each block
			for s in n1.block.successors: # for each node's block's branch target
				for n2 in self.nodes:
					if n2.block.instructions[0].lineNumber == s: # first instruction's line number
						n1.successors.append(n2)

		# add the end node
		for n in self.nodes: # for each node
			if n.successors == []:
				n.successors.append(self.endnode)
		self.nodes.append(self.endnode)

		# add the start node
		self.startnode.successors.append(self.nodes[0]) #add the start node
		self.nodes.insert(0,self.startnode)
		#self.startnode.successors.append(self.endnode)

		# populate the reverse CFG
		for n in self.nodes:
			for s in n.successors:
				s.predecessors.append(n)

		# for each node, populate domset
		self.genDomSets()

			

		# test...
		#for n in self.nodes:
		#	if n == self.startnode:
		#		print "startnode ->",
		#	elif n == self.endnode:
		#		print "endnode ->",
		#	else:
		#		print str(n.block.instructions[0].lineNumber) + " -> ",
		#	for s in n.successors:
		#		if s.block is not None:
		#			print str(s.block.instructions[0].lineNumber) + ", ",
		#		else:
		#			print "endnode",
		#	print ""
		#print ""
		#for n in self.nodes:
		#	if n == self.startnode:
		#		print "startnode <-",
		#	elif n == self.endnode:
		#		print "endnode <-",
		#	else:
		#		print str(n.block.instructions[0].lineNumber) + " <- ",
		#	for p in n.predecessors:
		#		if p.block is not None:
		#			print str(p.block.instructions[0].lineNumber) + ", ",
		#		else:
		#			print "startnode",
		#	print ""
		#print ""

	def makeCFG(self, blocks):
		# populate the nodes
		for b in blocks:
			temp = Node()
			temp.block = b
			self.nodes.append(temp)
#		self.help_makeCFG()
