#!/usr/bin/env python2.5

import sys
from globalvars import ISTREAM, OSTREAM
from instructions import Instruction
from blocks import Block
from functions import Function
from programs import Program


def cmdline(argv):
	opts = []
	backend = None
	for i in argv[1:]:
		if i.startswith('-opt='):
			for o in i[5:].split(','):
				opts.append(o)
		if i.startswith('-backend='):
			backend = i[9:]

	return opts, backend

def save_frontend(stream):
	isMain = False
	inst = None
	func = None
	prog = Program()

	for line in stream:
		inst = Instruction(line)
		if inst.opcode == 'entrypc':
			isMain = True
		if inst.opcode == 'enter':
			func = Function(isMain)
			isMain = False
		if func is not None:
			func.addInstruction(inst)
		if inst.opcode == 'ret':
			func.makeVariousInfo()
			prog.functions.append(func)
			func = None
	return prog

def frontend(stream):
	isMain = False
	inst = None
	func = None
	prog = Program()

	for line in stream:
		inst = Instruction()
		inst.makeInstruction(line)
		if inst.opcode == 'entrypc':
			isMain = True
		if inst.opcode == 'enter':
			func = Function(isMain)
			func.numLocals = (int(inst.ops[0].source) / 4)
			isMain = False
		if func is not None:
			func.instructions.append(inst)
		if inst.opcode == 'ret':
			#func.makeVariousInfo()
			prog.functions.append(func)
			func = None

	prog.buildBlocks()
	prog.buildCFG()
	return prog

def doAll(stream):
	opts, backend = cmdline(sys.argv)
	program = frontend(stream)

	for opt in opts:
		if opt == 'scp':
			pass # pass to Suriya's soln.
		elif opt == 'dce':
			pass # pass to Suriya's soln.
		elif opt == 'pre':
			pass # pass to Suriya's soln.
		elif opt == 'ssa':
			pass # pass to Suriya's soln.

	if backend == '3addr':
		pass
	elif backend == 'cfg':
		pass
	elif backend == 'c':
		pass
	elif backend == 'reg':
		#OSTREAM.write(" *  Backend:\treg\n\n")
		program.allocRegs()
		program.printReg()
	elif backend == 'ppc':
		program.collectGlobalVarSet()
		program.printPPC()

#ISTREAM = open('/u/dongli/home/TA/project/PPC/tb/s4.3addr.reg', 'r')
doAll(ISTREAM)
