
from globalvars import ISTREAM, OSTREAM
from blocks import Block
from instructions import Instruction
from operands import Operand
from cfg import CFG
from ssa import DF, IDF
from colorgraph import ColorGraph, ColorNode

from itertools import tee, izip

numGen = 1337000

def pairwise(iterable):
	a, b = tee(iterable)
	try:
		b.next()
	except StopIteration:
		pass
	return izip(a,b)

class Stacks:
	# a list of stacks!
	s = None

	def __init__(self):
		self.s = []
	
	def addVarToList(self, var):
		self.s.append([var,[]])
	
	def spush(self, var, subscript):
		for i in self.s:
			if i[0] is var:
				i[1].append(subscript)

	def top(self, var):
		for i in self.s:
			if i[0] is var:
				try:
					#print "top:\t" + var + ", " + str(len(i[1]))
					return i[1][len(i[1])-1]
				except IndexError:
					return 0

	def spop(self, var):
		for i in self.s:
			if i[0] is var:
				try:
					return i[1].pop()
					#print "pop:\t" + var + ", " + str(len(i[1]))
				except IndexError:
					return 0

class Counters:
	# a list of counters!
	c = None

	def __init__(self):
		self.c = []
	
	def addVarToList(self, var):
		self.c.append([var,0])
	
	def getCounter(self, var):
		for i in self.c:
			if i[0] is var:
				return i[1]
	
	def setCounter(self, var, x):
		for i in self.c:
			if i[0] is var:
				i[1] = x
	
	def incrementCounter(self, var):
		for i in self.c:
			if i[0] is var:
				i[1] += 1

class Function:
	isMain = None
	instructions = None
	cfg = None
	blocks = None
	stacks = None
	counters = None
	varSet = None
	tempSet = None
	numLocals = None

	creggraph = None
	reggraph = None
	minLocalVarOffset = 0
	maxParamOffset= 0

	def __init__(self, isMain):
		self.isMain = isMain
		self.instructions = []
		self.cfg = CFG()
		self.blocks = []
		self.stacks = Stacks()
		self.counters = Counters()
		self.varSet = set()
		self.tempSet = set()
		self.numLocals = 0
		self.creggraph = ColorGraph()
		self.reggraph = ColorGraph()
		self.globalVarSet= set()
		self.minLocalVarOffset = 0
		self.maxParamOffset= 0

	@staticmethod
	def genNum():
		global numGen
		numGen += 1
		return numGen

	def printInstructions(self):
		for i in self.instructions:
			str =  i.register + ": "
			str += i.expression
			OSTREAM.write('%s\n' % str)

	def print3addr(self):
		#OSTREAM.write('printing Function...\n')
		if self.isMain:
			s =  "    instr "
			s += str(self.blocks[0].instructions[0].lineNumber - 1) + ": "
			s += "entrypc"
			OSTREAM.write('%s\n' % s)
		#for b in self.blocks:
		#	b.print3addr()
		for n in self.cfg.nodes:
			b = n.block
			b.print3addr()

	def printReg(self):
		#OSTREAM.write('printing Function...\n')
		for n in self.cfg.nodes:
			b = n.block
			b.printReg()

	def collectGlobalVarSet(self):
		for n in self.cfg.nodes:
			b = n.block
			b.collectGlobalVarSet()
			self.globalVarSet = self.globalVarSet | b.globalVarSet
			if self.minLocalVarOffset > b.minLocalVarOffset:
				self.minLocalVarOffset = b.minLocalVarOffset
			if self.maxParamOffset< b.maxParamOffset:
				self.maxParamOffset= b.maxParamOffset
		OSTREAM.write('#The mininum local variable offset is: %d\n' % self.minLocalVarOffset )		
		OSTREAM.write('#The max parameter offset is: %d\n' % self.maxParamOffset)		
			
	def printPPC(self):
		if self.isMain:
			name = 'main'
		else:
#			name = 'label_' + str(self.cfg.nodes[0].successors[0].block.instructions[0].lineNumber)
			name = 'label_' + str(self.cfg.nodes[0].block.instructions[0].lineNumber)
		#OSTREAM.write('*   begin  Function\n')
		OSTREAM.write('\t.align 2\n')
		OSTREAM.write('\t.globl\t%s\n' % name)
		OSTREAM.write('\t.type\t%s, @function\n' % name)
		if self.isMain:
			OSTREAM.write('%s:\n' % name)

		for n in self.cfg.nodes:
			b = n.block
			if self.isMain:
				b.setInMain()
			b.set_maxParamOffset_global(self.maxParamOffset)
			b.printPPC()
			

		OSTREAM.write('\t.size %s' % name)
		OSTREAM.write(',.-%s\n' % name)

		#OSTREAM.write('*   end    Function\n')

	def buildBlocks(self):

		# compute leaders first
		leaders = set()
		leaders.add(self.instructions[0].lineNumber)
		for i in self.instructions:
			if i.opcode == 'br':
				leaders.update(i.successors)
			if i.opcode == 'blbc':
				leaders.update(i.successors)
			if i.opcode == 'blbs':
				leaders.update(i.successors)
			if i.opcode == 'call':
				leaders.update(i.successors)

		# create the Blocks
		curBlock = Block()
		for i in self.instructions:
			if i.lineNumber in leaders:
				curBlock = Block()
				self.blocks.append(curBlock)
			curBlock.instructions.append(i)

		# populate the successors' list
		for b in self.blocks:
			for s in b.instructions[len(b.instructions)-1].successors:
				b.successors.append(s)

		# test the data
		#for b in self.blocks:
		#	print "   block: " + str(b.instructions[0].lineNumber)
		#	for s in b.successors:
		#		print "successors: " + str(s)

	def nopInsn(self):
		nop = ' instr ' + str(Function.genNum()) + ': nop'
		nop_insn = Instruction()
		nop_insn.makeInstruction(nop)
		return nop_insn

	def buildCFG(self):
		self.cfg.makeCFG(self.blocks)
		self.cfg.startnode.block.instructions.append(self.nopInsn())
		self.cfg.endnode.block.instructions.append(self.nopInsn())

	def makeVariousInfo(self):
		"""computes various things"""
		self.makeCFG()

	def buildVarSet(self):
		l = set()
		for b in self.blocks:
			for i in b.instructions:
				if i.opcode == 'move':
					op = i.ops[1]
					if op.isVar:
						l.add(op.toString())
		self.varSet = l
		#return l
		

	def nodesWhichWriteToVar(self, var):
		l = set()
		for n in self.cfg.nodes:
			if not n.isStart and not n.isEnd:
				for i in n.block.instructions:
					if i.opcode == 'move':
						op = i.ops[1]
						if op.toString() == var:
							l.add(n)
		return l

	def insertPhi(self, var, node):

		phi_addy = Function.genNum()
		move_addy = Function.genNum()

		phi = Instruction()
		phi.lineNumber = phi_addy
		phi.opcode = 'phi'

		# add the right number of variables to phi
		for i in node.predecessors:
			phi.ops.append(Operand(var))

		node.block.instructions.insert(0, phi)

		# add the corresponding mov instruction
		move = Instruction()
		move.lineNumber = move_addy
		move.opcode = 'move'
		move.ops.append(Operand('(' + str(phi_addy) + ')'))
		move.ops.append(Operand(var))

		node.block.instructions.insert(1, move)

		# correct the pointers to this block
		for p in node.predecessors:
			last_insn = p.block.instructions[len(p.block.instructions)-1]
			if last_insn.opcode == 'br' or last_insn.opcode == 'blbs' or last_insn.opcode == 'blbc':
				last_insn.ops[len(last_insn.ops)-1] = Operand('['+ str(phi_addy) + ']')

	def phiNodes(self):
		varSet = self.buildVarSet()
		for var in self.varSet:
			hasAlready = set()
			everOnWorkList = set()
			workList = set()
			for x in self.nodesWhichWriteToVar(var):
				everOnWorkList.add(x)
				workList.add(x)
			while workList:
				x = workList.pop()
				for y in IDF(set([x])):
					if y not in hasAlready:
						self.insertPhi(var, y) # insert phi node for var at y
						hasAlready.add(y)
						if y not in everOnWorkList:
							everOnWorkList.add(y)
							workList.add(y)
	
	def genName(self, v):
		#print "v.var:\t" + v.var
		#print "v.source:\t" + v.source
		#print ""
		i = self.counters.getCounter(v.var)
		v.source = v.var + '$' + str(i)
		self.stacks.spush(v.var, i)
		self.counters.incrementCounter(v.var)


	def rename(self, x, visitedList): # block x
		if x in visitedList:
			return
		else:

			# first process phi nodes
			isPhiNode = False
			for i in x.block.instructions:
				if isPhiNode:
					assert i.opcode == 'move','broken phi node!'
					self.genName(i.ops[1])
					isPhiNode = False
				elif i.opcode == 'phi':
					isPhiNode = True

			# then process statements in block x
			for i in x.block.instructions:
				if i.opcode == 'move':
					if i.ops[0].isVar:
						i.ops[0].source = i.ops[0].var + '$' + str(self.stacks.top(i.ops[0].var))
					if i.ops[1].isVar:
						self.genName(i.ops[1])
				else:
					for op in i.ops:
						if op.isVar:
							s = str(self.stacks.top(op.var))
							if s == 'None':
								s = '0'
							op.source = op.var + '$' + s

			# then update any phi functions of successors of X
			#for y in x.predecessors:
			#for i in x.block.instructions:
			#	if i.opcode == 'phi':
			#		base = int(i.ops[0].source.split('$')[1])
			#		for op in i.ops[1:]:
			#			l = op.source.split('$')[0]
			#			op.source = l + '$' + str(base+1)
			#			base += 1
				
			# recurse
			for y in x.successors:
				self.rename(y, visitedList.union(set([x])))
			
			# pop variables defined in X when leaving
			isPhiNode = False
			for i in x.block.instructions:
				if isPhiNode:
					assert i.opcode == 'move','broken phi node!'
					#print "i.ops[2].var == " + i.ops[1].var
					self.stacks.spop(i.ops[1].var)
					isPhiNode = False
				elif i.opcode == 'phi':
					isPhiNode = True

	def renameVars(self):
		# initialize counters and stacks
		for v in self.varSet:
			self.stacks.addVarToList(v.split('#')[0])
			self.counters.addVarToList(v.split('#')[0])

		# call recursive rename
		self.rename(self.cfg.startnode, set())

	def ssa(self):
		self.phiNodes()
		self.renameVars()
	
	def pushTempToStack(self, temp):
		self.numLocals = self.numLocals + 1
		for b in self.blocks:
			for i in b.instructions:
				if i.opcode == 'enter':
					i.ops[0].source = str(self.numLocals * 4)

		name = 't' + temp
		offset = 0 - (self.numLocals * 4)
		var = name + '#' + str(offset)

		for b in self.blocks:
			for i in b.instructions:
				# if i consumes temp
				for op in i.ops:
					if op.source == '(' + temp + ')':
						pass

	def allocRegs(self):
		pass
