import re
from globalvars import OSTREAM
from operands import Operand

class Instruction:
	# EXAMPLE sourceCode:
	# 	instr 19: add a_base#32760 GP
	sourceCode = None
	lineNumber = None
	label = None
	opcode = None
	ops = None
	successors = None
	target = None
	targetReg = None
	hasPPCREG_OP = False
	forwardedOP = None
	globalVarSet = None
	parameterReg = 3
	minLocalVarOffset = 0
	minLocalVarOffset_g = 0
	maxParamOffset = 0
	maxParamOffset_g = 0

	cur_cmp_inst_op = ''
	inMainFunc = False
	def __init__(self):
		self.sourceCode = ''
		self.lineNumber = '0'
		self.opcode = ''
		self.ops = []
		self.successors = set()
		self.target = None
		self.targetReg = ''
		self.forwardedOP=None
		self.globalVarSet= set()
		
		Instruction.parameterReg = 3
		self.minLocalVarOffset = 0
		Instruction.minLocalVarOffset_g = 0
		Instruction.cur_cmp_inst_op = ''
		self.maxParamOffset= 0
		self.maxParamOffset_g= 0
		self.inMainFunc = False
	def setInMain(self):
		self.inMainFunc = True
	def set_maxParamOffset_global(self, gMax):
		self.maxParamOffset_g = gMax
	def makeInstruction(self, source):
		words = source.split()
		#assert words[0] == 'instr' # sanity check
		temp= words[0].split('~')
		if len(temp)>1:
			self.targetReg  = temp[0]
		self.sourceCode = source
		self.lineNumber = int(words[1].strip(':'))
		self.label = "label_" + str(self.lineNumber) + ":"
		self.opcode = words[2]
		self.ops = []
		for i in words[3:]:
			self.ops.append(Operand(i))
			if i.count('~') >0:
				self.hasPPCREG_OP = True
			match_base = re.match(r'(\w+_base)#([-\d]+)', i)
			if match_base is not None:
				self.globalVarSet.add(i.split('_base')[0])
			match_var = re.match(r'(\w+)#([-\d]+)',i)
			if match_var is not None:
				self.minLocalVarOffset = int(i.split('#')[1])
				if match_base is None:
					self.maxParamOffset= int(i.split('#')[1])	
				else:
					self.maxParamOffset= 0
				
		self.successors = set()

		if self.opcode == 'add':
			assert len(self.ops) == 2,'add len(self.ops) != 2'
			self.successors.add(self.lineNumber + 1)
			self.target = str(self.lineNumber)
		if self.opcode == 'sub':
			assert len(self.ops) == 2,'sub len(self.ops) != 2'
			self.successors.add(self.lineNumber + 1)
			self.target = str(self.lineNumber)
		if self.opcode == 'mul':
			assert len(self.ops) == 2,'mul len(self.ops) != 2'
			self.successors.add(self.lineNumber + 1)
			self.target = str(self.lineNumber)
		if self.opcode == 'div':
			assert len(self.ops) == 2,'div len(self.ops) != 2'
			self.successors.add(self.lineNumber + 1)
			self.target = str(self.lineNumber)
		if self.opcode == 'mod':
			assert len(self.ops) == 2,'mod len(self.ops) != 2'
			self.successors.add(self.lineNumber + 1)
			self.target = str(self.lineNumber)
		if self.opcode == 'neg':
			assert len(self.ops) == 1,'neg len(self.ops) != 1'
			self.successors.add(self.lineNumber + 1)
			self.target = str(self.lineNumber)
		if self.opcode == 'cmpeq':
			assert len(self.ops) == 2,'cmpeq len(self.ops) != 2'
			self.successors.add(self.lineNumber + 1)
			self.target = str(self.lineNumber)
		if self.opcode == 'cmple':
			assert len(self.ops) == 2,'cmple len(self.ops) != 2'
			self.successors.add(self.lineNumber + 1)
			self.target = str(self.lineNumber)
		if self.opcode == 'cmplt':
			assert len(self.ops) == 2,'cmplt len(self.ops) != 2'
			self.successors.add(self.lineNumber + 1)
			self.target = str(self.lineNumber)
		if self.opcode == 'param':
			assert len(self.ops) == 1,'param len(self.ops) != 1'
			self.successors.add(self.lineNumber + 1)
		if self.opcode == 'enter':
			assert len(self.ops) == 1,'enter len(self.ops) != 1'
			self.successors.add(self.lineNumber + 1)
		if self.opcode == 'ret':
			assert len(self.ops) == 1,'ret len(self.ops) != 1'
		if self.opcode == 'br':
			assert len(self.ops) == 1,'br len(self.ops) != 1'
			self.successors.add(int(self.ops[0].toString().strip('[]')))
		if self.opcode == 'blbc':
			assert len(self.ops) == 2,'blbc len(self.ops) != 2'
			self.successors.add(self.lineNumber + 1)
			self.successors.add(int(self.ops[1].toString().strip('[]')))
		if self.opcode == 'blbs':
			assert len(self.ops) == 2,'blbs len(self.ops) != 2'
			self.successors.add(self.lineNumber + 1)
			self.successors.add(int(self.ops[1].toString().strip('[]')))
		if self.opcode == 'call':
			assert len(self.ops) == 1,'call len(self.ops) != 1'
			self.successors.add(self.lineNumber + 1)
		if self.opcode == 'read':
			assert len(self.ops) == 0,'read len(self.ops) != 0'
			self.successors.add(self.lineNumber + 1)
			self.target = str(self.lineNumber)
		if self.opcode == 'write':
			assert len(self.ops) == 1,'write len(self.ops) != 1'
			assert self.hasPPCREG_OP, 'Only user register can be printed !, e.g. write r22~'
			self.successors.add(self.lineNumber + 1)
		if self.opcode == 'wrl':
			assert len(self.ops) == 0,'wrl len(self.ops) != 0'
			self.successors.add(self.lineNumber + 1)
		if self.opcode == 'load':
			assert len(self.ops) == 1,'load len(self.ops) != 1'
			self.successors.add(self.lineNumber + 1)
			self.target = str(self.lineNumber)
		if self.opcode == 'store':
			assert len(self.ops) == 2,'store len(self.ops) != 2'
			self.successors.add(self.lineNumber + 1)
		if self.opcode == 'move':
			assert len(self.ops) == 2,'move len(self.ops) != 2'
			self.successors.add(self.lineNumber + 1)
		if self.opcode == 'nop':
			assert len(self.ops) == 0,'nop len(self.ops) != 0'
			self.successors.add(self.lineNumber + 1)
		if self.opcode == 'entrypc':
			assert len(self.ops) == 0,'entrypc len(self.ops) != 0'
			self.successors.add(self.lineNumber + 1)
		if self.opcode == 'phi':
			assert len(self.ops) >= 0,'phi len(self.ops) < 0'
			self.successors.add(self.lineNumber + 1)
			self.target = str(self.lineNumber)

	def print3addr(self):
		#OSTREAM.write('printing Instruction...\n')
		s =  "    instr "
		s += str(self.lineNumber) + ": "
		s += self.opcode + " "
		for op in self.ops:
			s += op.toString() + " "
		s = s[:len(s)-1]
		OSTREAM.write("%s\n" % s)

	def printReg(self):
		#OSTREAM.write('printing Reg...\n')
		s = self.label + "\t"
		if self.opcode == 'add':
			s += str(self.target)
			s += " := "
			s += self.opcode
			for op in self.ops:
				s += " " + op.toReg()
		if self.opcode == 'sub':
			s += str(self.target)
			s += " := "
			s += self.opcode
			for op in self.ops:
				s += " " + op.toReg()
		if self.opcode == 'mul':
			s += str(self.target)
			s += " := "
			s += self.opcode
			for op in self.ops:
				s += " " + op.toReg()
		if self.opcode == 'div':
			s += str(self.target)
			s += " := "
			s += self.opcode
			for op in self.ops:
				s += " " + op.toReg()
		if self.opcode == 'mod':
			s += str(self.target)
			s += " := "
			s += self.opcode
			for op in self.ops:
				s += " " + op.toReg()
		if self.opcode == 'neg':
			s += str(self.target)
			s += " := "
			s += self.opcode
			for op in self.ops:
				s += " " + op.toReg()
		if self.opcode == 'cmpeq':
			s += str(self.target)
			s += " := "
			s += self.opcode
			for op in self.ops:
				s += " " + op.toReg()
		if self.opcode == 'cmple':
			s += str(self.target)
			s += " := "
			s += self.opcode
			for op in self.ops:
				s += " " + op.toReg()
		if self.opcode == 'cmplt':
			s += str(self.target)
			s += " := "
			s += self.opcode
			for op in self.ops:
				s += " " + op.toReg()
		if self.opcode == 'param':
			s += self.opcode
			for op in self.ops:
				s += " " + op.toReg()
		if self.opcode == 'enter':
			s += self.opcode
			for op in self.ops:
				s += " " + op.toReg()
		if self.opcode == 'ret':
			s += self.opcode
			for op in self.ops:
				s += " " + op.toReg()
		if self.opcode == 'br':
			s += self.opcode
			for op in self.ops:
				s += " " + op.toReg()
		if self.opcode == 'blbc':
			s += self.opcode
			for op in self.ops:
				s += " " + op.toReg()
		if self.opcode == 'blbs':
			s += self.opcode
			for op in self.ops:
				s += " " + op.toReg()
		if self.opcode == 'call':
			s += self.opcode
			for op in self.ops:
				s += " " + op.toReg()
		if self.opcode == 'read':
			s += str(self.target)
			s += " := "
			s += self.opcode
			for op in self.ops:
				s += " " + op.toReg()
		if self.opcode == 'write':
			s += self.opcode
			for op in self.ops:
				s += " " + op.toReg()
		if self.opcode == 'wrl':
			s += self.opcode
		if self.opcode == 'load':
			s += str(self.target)
			s += " := "
			s += self.opcode
			for op in self.ops:
				s += " " + op.toReg()
		if self.opcode == 'store':
			s += self.opcode
			for op in self.ops:
				s += " " + op.toReg()
		if self.opcode == 'move':
			s += self.opcode
			for op in self.ops:
				s += " " + op.toReg()
		if self.opcode == 'nop':
			s += self.opcode
		if self.opcode == 'entrypc':
			s += self.opcode
			for op in self.ops:
				s += " " + op.toReg()
		if self.opcode == 'phi':
			s += str(self.target)
			s += " := "
			s += self.opcode
			for op in self.ops:
				s += " " + op.toReg()
		OSTREAM.write("%s\n" % s)

	def hasConstant(self):
		result = False
		for op in self.ops:
			if op.isConstant:
				result = True
		return result

	def printPPC(self):
		#OSTREAM.write('printing Reg...\n')
		s = self.label + "\t"
		if self.opcode == 'add':
			if self.ops[1].source == 'GP':
				#s += 'lwz ' + self.target.strip('cr') + ',0(GP)\n\t\t'
				s +='lis 9, '
				s += self.ops[0].source.split('_base')[0]
				s +='@ha\n\t\t'
				
			elif self.hasConstant():
				s += 'addi '
				if self.targetReg is not None:
					s += self.targetReg.strip('r') + ','
				else:				
					s += self.target.strip('cr') + ',' 
				if self.ops[1].source == 'FP':
					s += self.ops[1].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g) + ','
					s += self.ops[0].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g)
				elif self.ops[1].source == 'GP':
					s += self.target.strip('cr') + ','
					s += self.ops[0].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g)
				else:
					s += self.ops[0].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g) + ','
					s += self.ops[1].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g)
			else:
				s += 'add '
				if self.targetReg is not None:
					s += self.targetReg.strip('r') + ','
				else:				
					s += self.target.strip('cr') + ',' 
				s += self.ops[0].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g) + ','
				s += self.ops[1].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g)
		if self.opcode == 'sub':
			if self.hasConstant():
				s += 'subi '
			else:
				s += 'sub '
			if self.targetReg is not None:
				s += self.targetReg.strip('r') + ','
			else:	
				s += self.target.strip('cr') + ','				
			s += self.ops[0].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g) + ','
			s += self.ops[1].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g)
		if self.opcode == 'mul':
			if self.hasConstant():
				s += 'mulli '
			else:
				s += 'mullw '
			if self.targetReg is not None:
				s += self.targetReg.strip('r') + ','
			else:				
				s += self.target.strip('cr') + ','
			s += self.ops[0].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g) + ','
			s += self.ops[1].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g)
		if self.opcode == 'div':
			if self.hasConstant():
				s += 'li  0, '
				s += self.ops[1].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g)
				s +='\n\t\t'
				s += 'divw '
				s += self.targetReg.strip('r') + ','
				s += self.ops[0].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g) + ', 0'
				s +='\n'
				OSTREAM.write("%s" % s)
				return
			s += 'divw '
			if self.targetReg is not None:
				s += self.targetReg.strip('r') + ','
			else:			
				s += self.target.strip('cr') + ','
			s += self.ops[0].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g) + ','
			s += self.ops[1].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g)
		if self.opcode == 'mod':
			# MUST BE EXPANDED
			# target = op0 % op1
			# a = b % c
			if self.hasConstant():
				s += 'li  0, '
				s += self.ops[1].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g)
				s += '\n\t\t'
				c = '0'
				a = self.targetReg.strip('r')
				b = self.ops[0].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g)
				s += 'divw ' + '4' + ',' + b + ',' + c
				s += '\n\t\t'
				s += 'mullw ' + '3' + ',' + '4' + ',' + c
				s += '\n\t\t'
				s += 'sub ' + a + ',' + b + ',' + '3'
				s += '\n'
				OSTREAM.write("%s" % s)
				return
			if self.targetReg is not None:
#				s += self.targetReg.strip('r') + ','
				a = self.targetReg.strip('r')
			else:
				a = self.target.strip('cr')
			b = self.ops[0].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g)
			c = self.ops[1].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g)
			#  a  = b / c
			#//  c = a * c
			#  r3 = a * c
			#  a  = b - r3
			s += 'divw ' + '4' + ',' + b + ',' + c
			s += '\n\t\t'
			s += 'mullw ' + '3' + ',' + '4' + ',' + c
			s += '\n\t\t'
			s += 'sub ' + a + ',' + b + ',' + '3'
			#s += self.ops[1].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g)
		if self.opcode == 'neg':
			s += 'neg '
			if self.targetReg is not None:
				s += self.targetReg.strip('r') + ','
			else:
				s += self.target.strip('cr') + ','
			s += self.ops[0].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g)
		if self.opcode == 'cmpeq':
			Instruction.cur_cmp_inst_op = 'cmpeq'
			if self.ops[1].isConstant:
				s += 'cmpwi '
			else:
				s += 'cmpw '
#			s += self.target.strip('cr') + ','
			s += '0, '
			s += self.ops[0].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g) + ','
			s += self.ops[1].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g)
			s +='\n\t\t'
#			s += 'bc 4, 2, '
			OSTREAM.write("%s" % s)
			return
		if self.opcode == 'cmple':
			Instruction.cur_cmp_inst_op = 'cmple'
			if self.ops[1].isConstant:
				s += 'cmpwi '
			else:
				s += 'cmpw '
#			s += self.target.strip('cr') + ','
			s += '0, '
			s += self.ops[0].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g) + ','
			s += self.ops[1].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g)
			s +='\n\t\t'
#			s += 'bc 0xC, 1, '
			OSTREAM.write("%s" % s)
			return
		if self.opcode == 'cmplt':
			Instruction.cur_cmp_inst_op = 'cmplt'
			if self.ops[1].isConstant:
				s += 'cmpwi '
			else:
				s += 'cmpw '
#			s += self.target.strip('cr') + ','
			s += '0, '
			s += self.ops[0].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g) + ','
			s += self.ops[1].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g)
			s +='\n\t\t'
#			s += 'bc 4, 0, '
			OSTREAM.write("%s" % s)
			return
		if self.opcode == 'param':
			assert Instruction.parameterReg <=10, 'The number of parameters are bigger than 8, not allowed in this lab'
			#s += 'stw '
			#s += value
			#s += dest
			if self.ops[0].isConstant:
				s += 'li '
			else:
				s += 'mr '
			s += str(Instruction.parameterReg)
			s += ',' + self.ops[0].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g)
			Instruction.parameterReg +=1
			
		if self.opcode == 'enter':
			Instruction.parameterReg = 3
			Instruction.minLocalVarOffset_g = int(self.ops[0].source)
			if False:
#			if self.inMainFunc:
				s += 'stwu 1,-48(1)'
				s += '\n\t\t'
				s += 'mflr 0'
				s += '\n\t\t'
				s += 'stw 31,44(1)'
				s += '\n\t\t'
				s += 'stw 0,52(1)'
				s += '\n\t\t'
				s += '\n\t\t'
				OSTREAM.write('#This is the Main Function \n')
			else:
				s += 'stw     12,-80(1) \n\t\t' 
				s += 'stw     13,-76(1) \n\t\t' 
				s += 'stw     14,-72(1) \n\t\t' 
				s += 'stw     15,-68(1) \n\t\t' 
				s += 'stw     16,-64(1) \n\t\t' 
				s += 'stw     17,-60(1) \n\t\t' 
				s += 'stw     18,-56(1) \n\t\t' 
				s += 'stw     19,-52(1) \n\t\t' 
				s += 'stw     20,-48(1) \n\t\t' 
				s += 'stw     21,-44(1) \n\t\t' 
				s += 'stw     22,-40(1) \n\t\t' 
				s += 'stw     23,-36(1) \n\t\t' 
				s += 'stw     24,-32(1) \n\t\t' 
				s += 'stw     25,-28(1) \n\t\t' 
				s += 'stw     26,-24(1) \n\t\t' 
				s += 'stw     27,-20(1) \n\t\t' 
				s += 'stw     28,-16(1) \n\t\t' 
				s += 'stw     29,-12(1) \n\t\t' 
				s += 'stw     30, -8(1) \n\t\t' 
				s += 'stw     31, -4(1) \n\t\t' 

				next_address = -80 -int(self.ops[0].source) -4
				s += 'stw     3, '  + str(next_address)+ '(1) \n\t\t'
				next_address = next_address - 4
				s += 'stw     4, '  + str(next_address)+ '(1) \n\t\t'
				next_address = next_address - 4
				s += 'stw     5, '  + str(next_address)+ '(1) \n\t\t'
				next_address = next_address - 4
				s += 'stw     6, '  + str(next_address)+ '(1) \n\t\t'
				next_address = next_address - 4
				s += 'stw     7, '  + str(next_address)+ '(1) \n\t\t'
				next_address = next_address - 4
				s += 'stw     8, '  + str(next_address)+ '(1) \n\t\t'
				next_address = next_address - 4
				s += 'stw     9, '  + str(next_address)+ '(1) \n\t\t'
				next_address = next_address - 4
				s += 'stw     10, ' + str(next_address)+ '(1) \n\t\t'
				
				next_address = next_address - 4
				s += 'stw     2, ' + str(next_address)+ '(1) \n\t\t'

				next_address = next_address - 12
				s += ''
				s += 'mflr  0\n\t\t'
				s += 'stw     0, ' + str(next_address)+ '(1) \n\t\t'

				next_address = next_address - 4
				s += ''
				s += 'mfcr  0\n\t\t'
				s += 'stw     0, ' + str(next_address)+ '(1) \n\t\t'

				next_address = next_address - 4
				s += ''
				s += 'stwu     1, ' + str(next_address)+ '(1) \n\t\t'	
				

			#s += 'stwu '
			#s += '1,' + str((-4 * int(self.ops[0].source)) - 32) + '(1)'
		if self.opcode == 'ret':
				frame_size = 136-self.minLocalVarOffset_g
				s += 'lwz 11,0(1) \n\t\t'
				s += 'mr 1, 11 \n\t\t'

				s += 'lwz     12,-80(1) \n\t\t' 
				s += 'lwz     13,-76(1) \n\t\t' 
				s += 'lwz     14,-72(1) \n\t\t' 
				s += 'lwz     15,-68(1) \n\t\t' 
				s += 'lwz     16,-64(1) \n\t\t' 
				s += 'lwz     17,-60(1) \n\t\t' 
				s += 'lwz     18,-56(1) \n\t\t' 
				s += 'lwz     19,-52(1) \n\t\t' 
				s += 'lwz     20,-48(1) \n\t\t' 
				s += 'lwz     21,-44(1) \n\t\t' 
				s += 'lwz     22,-40(1) \n\t\t' 
				s += 'lwz     23,-36(1) \n\t\t' 
				s += 'lwz     24,-32(1) \n\t\t' 
				s += 'lwz     25,-28(1) \n\t\t' 
				s += 'lwz     26,-24(1) \n\t\t' 
				s += 'lwz     27,-20(1) \n\t\t' 
				s += 'lwz     28,-16(1) \n\t\t' 
				s += 'lwz     29,-12(1) \n\t\t' 
				s += 'lwz     30, -8(1) \n\t\t' 
				s += 'lwz     31, -4(1) \n\t\t' 

				next_address = -80 -self.minLocalVarOffset_g -4
			  	s += 'lwz     3, '  + str(next_address)+ '(1) \n\t\t'
			        next_address = next_address - 4
			  	s += 'lwz     4, '  + str(next_address)+ '(1) \n\t\t'
			        next_address = next_address - 4
			  	s += 'lwz     5, '  + str(next_address)+ '(1) \n\t\t'
			        next_address = next_address - 4
			  	s += 'lwz     6, '  + str(next_address)+ '(1) \n\t\t'
			        next_address = next_address - 4
			  	s += 'lwz     7, '  + str(next_address)+ '(1) \n\t\t'
			        next_address = next_address - 4
			  	s += 'lwz     8, '  + str(next_address)+ '(1) \n\t\t'
			        next_address = next_address - 4
			  	s += 'lwz     9, '  + str(next_address)+ '(1) \n\t\t'
			        next_address = next_address - 4
			  	s += 'lwz     10, ' + str(next_address)+ '(1) \n\t\t'
				
				next_address = next_address - 4
				s += 'lwz     2, ' + str(next_address)+ '(1) \n\t\t'

				next_address = next_address - 12
				s += ''
				s += 'lwz     0, ' + str(next_address)+ '(1) \n\t\t'
				s += 'mtlr  0\n\t\t'

				next_address = next_address - 4
				s += ''
				s += 'lwz     0, ' + str(next_address)+ '(1) \n\t\t'
				s += 'mtcrf 255, 0\n\t\t'
				s += 'blr'
				Instruction.minLocalVarOffset_g = 0
		if self.opcode == 'br':
			s += 'b '
			s += self.ops[0].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g)
		if self.opcode == 'blbc':
			if Instruction.cur_cmp_inst_op == 'cmpeq':
				s += 'bc 0x4, 2, '#not eq,  br
			if Instruction.cur_cmp_inst_op == 'cmplt':
				s += 'bc 0x4, 0, '#not eq,  br
			if Instruction.cur_cmp_inst_op == 'cmple':
				s += 'bc 0xC, 1, '#not eq,  br
			s += self.ops[1].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g)
			OSTREAM.write("%s\n" % s)
			return
#			s += 'bc '
##			s += self.ops[0].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g)
#			s += ' 0x7'
#			s += ',0,'
#			s += self.ops[1].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g) 
		if self.opcode == 'blbs':
			if Instruction.cur_cmp_inst_op == 'cmple':
				s += 'bc 0x4, 1, '#not eq,  br
			if Instruction.cur_cmp_inst_op == 'cmplt':
				s += 'bc 0xC, 0, '#not eq,  br
			if Instruction.cur_cmp_inst_op == 'cmpeq':
				s += 'bc 0xC, 2, '#not eq,  br
			s += self.ops[1].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g)
#			s = self.ops[1].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g)
			OSTREAM.write("%s\n" % s)
			return
#			s += 'bc '
##			s += self.ops[0].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g)
#			s += ' 0xc'			
#			s += ',1,'
#			s += self.ops[1].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g) 
		if self.opcode == 'call':
			Instruction.parameterReg = 3
			s += 'bl '
			s += self.ops[0].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g)
		if self.opcode == 'read':
			s += 'bl ReadLong'
		if self.opcode == 'write':
			s += 'lis 9,.LC0@ha'
			s += '\n\t\t'
			s += 'la 3,.LC0@l(9)'
			s += '\n\t\t'
			s += 'mr 4, '
			s += self.ops[0].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g)
			s += '\n\t\t'
			s += 'crxor 6,6,6'
			s += '\n\t\t'
			s += 'bl printf'
			s += '\n\t\t'
			s += 'lwz 11,0(1)'
			s += '\n\t\t'
			s += 'lwz 0,4(11)'
			s += '\n\t\t'
			s += 'mtlr 0'
			s += '\n\t\t'
			#s += 'lwz 31,-4(11)'
			#s += '\n\t\t'
			#s += 'mr 1,11'
			#s += '\n\t\t'

			#s += 'bl WriteLong'
		if self.opcode == 'wrl':
			s += 'bl WriteLine'
		if self.opcode == 'load':
			if self.ops[0].isBase:
				s+=  'lwz 0, '
				s += self.ops[0].source.split('_base')[0]
				s +='@l(9)\n\t\t'
				
				s +='mr '
				s += self.targetReg.strip('r')
				s += ', 0'
								
				OSTREAM.write("%s\n" % s)
				return
			
			s += 'lwz '
			if self.targetReg is not None:
				s += self.targetReg.strip('r') + ','
				s += self.ops[0].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g) + '(1)'
			else:			
				s += self.target.strip('cr') + ','
				s += '0(' + self.ops[0].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g) + ')'
		if self.opcode == 'store':
			
			if self.ops[1].isBase:
				if self.ops[0].isConstant:
					s += 'li 0, 0'
					s += '\n\t\t'
					s +='ori 0, 0, '
					s += self.ops[0].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g)
				elif self.ops[0].isReg:
					s += 'mr 0, '
					s += self.ops[0].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g)
				s +='\n\t\t'
				s+=  'stw 0, '
				s += self.ops[1].source.split('_base')[0]
				s +='@l(9)\n\t\t'
				OSTREAM.write("%s\n" % s)
				return
					
					
			if self.ops[1].ppcReg is not None:
				if self.ops[0].isPPCREG:
					s += 'stw '
					s += self.ops[0].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g) + ','
					s += self.ops[1].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g) + '(1)'
				elif self.ops[0].isConstant:
					#s += 'li 0, '
					s += 'li 0, 0'
					s += '\n\t\t'
					s +='ori 0, 0, '
					s += self.ops[0].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g)
					s += '\n\t\t'
					s +='stw 0, '
					s +=self.ops[1].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g) +'(1)'
			else:
				s += 'stw '
				s += self.ops[0].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g) + ','
				s += '0(' + self.ops[1].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g) + ')'
		if self.opcode == 'move':
			if self.hasPPCREG_OP:
				if self.ops[0].isConstant:
					s += 'li '
				else:
					s += 'mr '
				s += self.ops[1].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g) + ', '
				s += self.ops[0].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g)
			else:
				#s += 'li 0, '
				s += 'li 0, 0'
				s += '\n\t\t'
				s +='ori 0, 0, '
				s += self.ops[0].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g)
				s += '\n\t\t'
				s +='stw 0, '
				s +=self.ops[1].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g) +'(1)'
			#s += 'li '
			#s += self.ops[1].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g) + ','
			#s += self.ops[0].toPPC(self.maxParamOffset_g, self.minLocalVarOffset_g)

		if self.opcode == 'nop':
			pass # no need to print one! :)
		if self.opcode == 'entrypc':
			pass # no need to print one! :)
		if self.opcode == 'phi':
			assert False,'Phi instructions must be removed for backend=ppc'

		OSTREAM.write("%s\n" % s)
